(* file bogenLexer.mll *)
{
(** Lexikalische Analyse der Eingabegrammatik. Automatisch generiert per
 *  [ocamllex] aus der Datei [bogenLexer.mll].
 *
 *  Die verwendeten Tokentypen sind im Modul BogenParser definiert.
 *)
open BogenParser
open Printf
}

rule token = parse
    (* Leerzeichen, Tabulatoren und Carriage returns werden ignoriert: *)
    [' ' '\t' '\r']  { token lexbuf }
  | (* Bei Zeilenumbruechen wird der Wert im Feld
       lexbuf.Lexing.lex_curr_p.pos_lnum inkremetiert, ansonsten werden
       auch sie ignoriert:
     *)
    '\n' { Lexing.new_line lexbuf; token lexbuf }
  | (* Kommentare: *)
    "/*" { comments 0 lexbuf }
  | (* Der Header enthaelt C-Programmcode, umschlossen von %{ und %}, und
     * soll den Scanner ungeprueft passieren. Zurueckgeliefert wird ein
     * Integerpaar fuer Start- und Endposition des Header-Codes ohne die
     * Anfangs- und Endmarken.
     *)
    "%{" ( [^ '%'] | ('%'[^ '}']) )* "%}"
    { let h_start = (Lexing.lexeme_start lexbuf) + 2
      in let h_end = (Lexing.lexeme_end lexbuf) - 2
      in
        HEADER (h_start, h_end)
    }
  | "%token"    { DECL_TOKEN }
  | "%start"    { DECL_START } 
  | "%%"        { DELIMITER (Lexing.lexeme_end lexbuf) }
  | ['a'-'z' 'A'-'Z']['a'-'z' 'A'-'Z' '0'-'9' '_']* as id
    { IDENTIFIER id }
  | ':' { COLON }
  | '|' { PIPE }
  | ';' { SEMICOLON }
  | (* Semantische Aktionen fuer Regeln: *)
    '{' ([^ '}']* as action) '}'  { SEMANTIC_ACTION action }
  | _ as c
    { print_endline ("Unrecognized character: "
				^ (Char.escaped c) ^ " in line " 
				^ (string_of_int lexbuf.Lexing.lex_curr_p.Lexing.pos_lnum) );
      token lexbuf
    }
  | eof { EOF }

(* Kommentare, auch verschachtelt, werden erkannt und ignoriert. *)
and comments level = parse
    "*/" { if level = 0 then token lexbuf else comments (level-1) lexbuf }
  | "/*" { comments (level+1) lexbuf }
  | '\n' { Lexing.new_line lexbuf; comments level lexbuf }
  | _ { comments level lexbuf }
  | eof { raise End_of_file }
