/**************************************************************************
 * file bogenParser.mly                                                   *
 * Spezifikation des Eingabeformats fuer BoGenLR                          *
 **************************************************************************/

/* Das Format fuer Grammatikspezifikationen lehnt sich an die Syntax der
 * Eingabedateien fuer Yacc und Bison an.
 *
 * Zur Vereinfachung und zwecks Konzentration auf das Wesentliche werden in
 * dieser Version allerdings einige Konzepte weggelassen, die gaengige
 * Parsergeneratoren wie Yacc/Bison unterstuetzen. Dazu gehoeren:
 * - Deklarationen fuer die Assoziativitaet und Praezedenz von Token:
 *   %left, %right, %nonassoc
 * - Typisierung der Token und Nichtterminale, um den Typ des
 *   Rueckgabewertes der assoziierten semantischen Aktionen festzulegen.
 *   Semantische Aktionen koennen zwar angegeben werden, werden aber
 *   ignoriert. Ausserdem duerfen Sie nicht mitten in einer Regeldefinition
 *   stehen, sondern nur am Ende, und sie duerfen keine geschweiften
 *   Klammern enthalten.
 * - Definition von mehr als einem Startsymbol. Dies stellt aber keine echte
 *   Einschraenkung dar, weil eine Grammatik mit mehreren Startsymbolen
 *   durch eine zusaetzliche Produktion leicht umgeformt werden kann.
 */

%{
  (** Syntaxanalyse der Eingabegrammatik. Automatisch generiert via
   *  [ocamlyacc] aus der Spezifikationsdatei [bogenParser.mly].
   *)

  open BogenGrammar
	
  (** Initialisierung einer Hashtabelle fuer die Terminalsymbole. Die
   *  Symbole werden in den Schluesselfeldern abgelegt, in den Datenfeldern
   *  wird jedem Symbol eine eindeutige Nummer zugewiesen.
   *)
  let token_tbl = Hashtbl.create 7
	
  (** Initialisierung einer Hashtabelle fuer die Nichtterminalsymbole. Die
	 *	Symbole werden in den Schluesselfeldern abgelegt. In den Datenfeldern
   *  werden bei der Generierung des Resultatparsers eine durchgehende
   *  Nummerierung sowie die FIRST_k-Mengen abgelegt.
   *)
  let nonterminal_tbl = Hashtbl.create 7
	
  (** Initialisierung des Startsymbols als leere Referenz. *)
  let startsym = ref None

  (** Die Tokens werden in der Reihenfolge ihrer Entdeckung fortlaufend
   *  durchnummeriert, beginnend bei 258.
   *)
  let token_number = ref 257
  
  (** Die Nichtterminalsymbole werden in der Reihenfolge ihrer Entdeckung
   *  fortlaufend durchnummeriert, beginnend bei 1.
   *)
  let nonterminal_number = ref 0

  (** Die Regeln werden in der Reihenfolge ihrer Entdeckung
   *  fortlaufen durchnummeriert, beginnend bei 0.
   *)
  let rule_number = ref 0
%}


/**************************************************************************
 * Tokendefinitionen:                                                     *
 **************************************************************************/

/* Der Lexer liefert fuer den Header ein Paar aus Start- und
 * Endposition:
 */
%token <int * int> HEADER

/* Tokens fuer die Tokentyp- und Startsymboldeklarationen %token und
 * %start:
 */
%token DECL_TOKEN DECL_START

/* Token fuer doppelte Prozentzeichen als Bereichstrenner. Der
 * Rueckgabewert ist die aktuelle Position in der Eingabedatei und wird
 * benoetigt, um die Position von evtl. vorhandenem Trailercode in der
 * Eingabedatei zu bestimmen.
 */
%token <int> DELIMITER

/* Gueltige Bezeichner fuer Terminal- und Nichtterminalsymbole: */
%token <string> IDENTIFIER

/* Trennzeichen:
 * Der Doppelpunkt trennt Kopf und Rumpf von Grammatikregeln;
 * ein senkrechter Strich trennt Alternativen in Grammatikregeln;
 * das Semikolon schließt eine Grammatikregel ab;
 * das Token EOF markiert das Ende der Eingabedatei.
 */
%token COLON PIPE SEMICOLON EOF

/* Semantische Aktionen zu jeder Grammatikregel sind in geschweifte
 * Klammern eingeschlossen.
 */
%token <string> SEMANTIC_ACTION


/**************************************************************************
 * Startsymbol und spezielle Nichtterminal-Typen:                         *
 **************************************************************************/

/* Wenn die Eingabegrammatik erfolgreich geparst wurde, wird eine interne
 * Grammatikdarstellung vom Typ BogenGrammar.t zurueckgegeben.
 */
%type <BogenGrammar.t> grammar
%start grammar

/* end_of_grammar liefert die Position, an der der Regelteil endet und der
 * Trailer beginnt. Ein Wert von 0 bedeutet, dass die
 * Grammatikspezifikation keinen Trailer enthaelt.
 */
%type <int> end_of_grammar


%%


/**************************************************************************
 * Regeln:                                                                *
 **************************************************************************/

grammar :
    header declarations rules end_of_grammar
    { let (hs, he) = $1
      in let s x = match x
      with
				Some y -> y
			| None -> ""
			in let
        result : BogenGrammar.t = {
					nonterminals = nonterminal_tbl;
					terminals = token_tbl;
					productions = $3;
					startsymbol = (s !startsym);
					header_start = hs;
					header_end = he;
					trailer_start = $4
				}
			in
        result
    }
  ;


/* Der gesamte Header passiert den Parser ungeprueft. Resultatwert ist
 * Paar von Integers fuer Start- und Endposition des Headercodes in der
 * Eingabegrammatik. Er darf auch fehlen, dann wird das Paar (0, 0)
 * zurueckgeliefert.
 */
header :
    { (0, 0) }
  | HEADER
    { $1 }
  ;


/* Der Deklarationsteil besteht aus einer Folge von Token- und
 * Startsymboldeklarationen, gefolgt vom Trennertoken %%. Er darf auch
 * leer sein, aber dann muss dem Regelteil mindestens das Trennertoken %%
 * vorangehen.
 * Da die Tokens direkt in eine globale Hashtabelle eingetragen werden und
 * eine evtl. vorhandene Startsymboldeklaration als String-Referenz
 * gespeichert wird, werden Rueckgabewerte hier nicht verwendet.
 */
declarations :
    decls DELIMITER { () }
  ;
  

decls :
    { () }
  | decls tokendecl { () }
  | decls startdecl { () }
  ;
 

tokendecl :
    DECL_TOKEN tokenids { () }
  ;


/* In einer Grammatik ist hoechstens eine Startsymboldeklaration erlaubt.
 */
startdecl :
    DECL_START IDENTIFIER
    { begin
        if !startsym = None
        then startsym := Some $2
        else raise BogenGrammar.Startsymbol_already_exists;
      end
    }
  ;


/* Jedes deklarierte Token wird mit einer eindeutigen Nummer versehen und
 * in der Hashtabelle token_tbl eingetragen.
 * Sollte ein Token vorher schon einmal deklariert worden sein, liegt ein
 * Fehler in der Eingabegrammatik vor.
 */
tokenids :
    more_tokenids IDENTIFIER
    { begin
        if Hashtbl.mem token_tbl $2
        then raise (BogenGrammar.Token_already_exists $2)
        else begin
          token_number := !token_number + 1;
          Hashtbl.add token_tbl $2 !token_number;
        end;
      end
    }
  ;
  

more_tokenids :
    { () }
  | more_tokenids IDENTIFIER
    { begin
        if Hashtbl.mem token_tbl $2
        then raise (BogenGrammar.Token_already_exists $2)
        else begin
          token_number := !token_number + 1;
          Hashtbl.add token_tbl $2 !token_number;
        end;
      end
    }
  ;


/* Eine Grammatik ohne Regeln ist unnuetz, also muss der Regelteil
 * mindestens eine erste Regel enthalten. Weitere Regeln koennen folgen.
 * Rueckgabewert von first_rule und more_rules sind jeweils eine Liste von
 * Einzelregeln. Rueckgabewert von rules ist deren Konkatenation, ergaenzt
 * um eine durchlaufende, bei 0 beginnende Nummerierung der Regeln.
 */
rules :
    first_rule more_rules
    { List.map
        ( fun (lhs, rhs) -> begin
            rule_number := !rule_number + 1;
            (!rule_number, lhs, rhs);
          end;
        )
        ($1 @ $2 )
    }
  ;


/* Wenn im Deklarationsteil noch kein Startsymbol der Grammatik
 * festgelegt wurde, wird das Nichtterminalsymbol auf der linken Seite
 * der ersten Regel als Startsymbol angenommen.
 * Fuer jede Regel wird das Nichtterminalsymbol auf der linken Seite in
 * die Hashtabelle nonterminal_tbl aufgenommen und mit einer eindeutigen
 * Nummer versehen, falls es nicht schon enthalten ist.
 * Die Liste der Alternativen [rhs1; rhs2; ...] fuer die betrachtete Regel
 * wird aufgespalten. Rueckgabewert ist eine Liste der erzeugten
 * Einzelregeln in der Form
 * [ (nonterminal, rhs1); (nonterminal, rhs2); ... ]
 */
first_rule :
    nonterminal COLON alternatives SEMICOLON
    { begin
        if !startsym = None
        then startsym := Some $1;

        if not (Hashtbl.mem nonterminal_tbl $1)
        then begin
          nonterminal_number := !nonterminal_number + 1;
          Hashtbl.add nonterminal_tbl $1
            {number = !nonterminal_number; items = []; first_k = []};
        end;
      end;
      List.map (fun x -> ($1, x)) $3
    }
  ;


/* Sowohl more_rules als auch rule repraesentieren Listen von Regeln.
 * Rueckgabewert ist eine Konkatenation dieser Listen.
 */
more_rules :
    more_rules rule { $1 @ $2 }
  | { [] }
  ;


/* Fuer jede Regel wird das Nichtterminalsymbol auf der linken Seite in
 * die Hashtabelle nonterminal_tbl aufgenommen und mit einer eindeutigen
 * Nummer versehen, falls es nicht schon enthalten ist.
 * Die Liste der Alternativen [rhs1; rhs2; ...] fuer die betrachtete Regel
 * wird aufgespalten. Rueckgabewert ist eine Liste der erzeugten
 * Einzelregeln in der Form
 * [ (nonterminal, rhs1); (nonterminal, rhs2); ... ]
 */
rule :
    nonterminal COLON alternatives SEMICOLON
    { if not (Hashtbl.mem nonterminal_tbl $1)
      then begin
        nonterminal_number := !nonterminal_number + 1;
        Hashtbl.add nonterminal_tbl $1
          {number = !nonterminal_number; items = []; first_k = []};
      end;
      List.map (fun x -> ($1, x)) $3
    }
  ;


/* Ein Symbol, das auf der linken Seite einer Regel steht, darf nicht zuvor
 * als Token deklariert worden sein.
 */
nonterminal :
    IDENTIFIER
    { if Hashtbl.mem token_tbl $1
      then raise (Token_as_lhs $1)
      else $1
    }
  ;


/* Jede Regel hat mindestens eine erste Alternative.
 * Rueckgabewert von first_alternative ist eine Liste von Symbolen.
 * Rueckgabewert von more_alternatives ist eine Liste von Alternativen,
 * also eine Liste aus Listen von Symbolen.
 * Der Rueckgabewert von alternatives fuegt die erste Alternative vor die
 * Liste der restlichen Alternativen.
 */
alternatives :
    first_alternative more_alternatives { $1 :: $2 }
  ;


/* Die erste Alternative darf nicht mit einem "|" beginnen.
 * Semantische Aktionen fuer die einzelnen Alternativen, eingeschlossen
 * in { und }, sind vorgesehen, werden aber in der aktuellen Version
 * des Generators nicht ausgewertet. Im Gegensatz zu Bison sind sie nur
 * am Ende jeder Alternative erlaubt.
 */
first_alternative :
    { [] (* Alternative, die nur aus dem leeren Wort besteht *) }
  | symbols symbol SEMANTIC_ACTION { $1 @ [$2] }
  ;


more_alternatives :
    { [] (* Leere Liste von Alternativen. *)}
  | more_alternatives alternative { $1 @ [$2] }
  ; 


alternative :
    PIPE SEMANTIC_ACTION
    { [] (* Alternative, die nur aus dem leeren Wort besteht *) }
  | PIPE symbols symbol SEMANTIC_ACTION { $2 @ [$3] }
  ;              


symbols :
    { [] }
  | symbols symbol { $1 @ [$2] }
  ;


/* Ein Symbol auf der rechten Seite einer Regel, das nicht im
 * Deklarationsteil als Terminalsymbol ausgewiesen wurde, muss ein
 * Nichtterminalsymbol sein, dessen Regel spaeter im Regelteil noch folgt.
 */
symbol :
    IDENTIFIER
    { if (Hashtbl.mem token_tbl $1) then
        Terminal $1
      else
        begin
          if not (Hashtbl.mem nonterminal_tbl $1)
          then begin
            nonterminal_number := !nonterminal_number + 1;
            Hashtbl.replace nonterminal_tbl $1
              {number = !nonterminal_number; items = []; first_k = []};
          end;
          Nonterminal $1
        end;
    }
  ;
  
/* Die Grammatik ist fertig eingelesen, wenn entweder das Dateiende
 * erreicht wurde oder das im Regelteil das Trennzeichen "%%" gelesen wird,
 * das den Trailercode einleitet.
 */
end_of_grammar :
    EOF { 0 }
  | DELIMITER { $1 }
  ;
