/*	file mmax.c
 *
 *	Programmcode, der als Shared Library bersetzt werden sollte,
 *	die bei Bedarf geladen wird, um den Speicherbedarf eines Programms zu
 *	bestimmen. Funktioniert aufgrund der Verwendung der kvm-Library nur
 *	unter NetBSD!
 *
 *	bersetzen: gcc -fpic -shared -c -o mmax.o mmax.c
 *	Linken: ld -lkvm -G -o mmax.so mmax.o
 *	Aufruf: env LD_PRELOAD=`pwd`/mmax.so <programm>
 */

#include <dlfcn.h>
#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>

/* Fr kvm: */
#include <fcntl.h>
#include <kvm.h>
#include <sys/param.h>
#include <sys/sysctl.h>

#include <inttypes.h>


static void    *lib = NULL;
/* static size_t  maxalloc = 0U; */
int64_t max_mem = 0;

union malloc {
    void *(*f)(size_t);
    void *ptr;
};

union calloc {
    void *(*f)(size_t, size_t);
    void *ptr;
};

union realloc {
    void *(*f)(void *, size_t);
    void *ptr;
};

static union malloc   oldmalloc;
static union calloc   oldcalloc;
static union realloc  oldrealloc;


static kvm_t *kd = NULL; // kvm
static int mypid;


static
void
print_stuff(void)
{
    fprintf(stderr, "pid: %d, max. memory used: %" PRId64 " kB\n",
                    mypid, max_mem * 4);
/*    fprintf(stderr, "maximum allocated: %zu\n", maxalloc); */
    kvm_close(kd);
}


static bool in_update_stats = false;

static
void
update_stats(void)
{
    int cnt; // kvm_getproc2
    struct kinfo_proc2 *stats;


    if (kd != NULL && !in_update_stats) {
        in_update_stats = true;
        stats = kvm_getproc2(kd, KERN_PROC_PID, mypid,
                             sizeof(struct kinfo_proc2), &cnt);
        if (stats == NULL) {
            perror("kvm_getproc2()");
            abort();
        }
        if (max_mem < stats->p_vm_msize)
            max_mem = stats->p_vm_msize;
        in_update_stats = false;
    }
}


static
void
initialize(void)
{
    lib = dlopen("libc.so", RTLD_LAZY);
    if (lib == NULL) {
        fprintf(stderr, "cannot open \"libc.so\"...\n");
        abort();
    }

    oldmalloc.ptr = dlsym(lib, "malloc");
    if (oldmalloc.ptr == NULL) {
        fprintf(stderr, "cannot get \"malloc()\"...\n");
        abort();
    }

    oldcalloc.ptr = dlsym(lib, "calloc");
    if (oldcalloc.ptr == NULL) {
        fprintf(stderr, "cannot get \"calloc()\"...\n");
        abort();
    }

    oldrealloc.ptr = dlsym(lib, "realloc");
    if (oldrealloc.ptr == NULL) {
        fprintf(stderr, "cannot get \"realloc()\"...\n");
        abort();
    }

    if (atexit(print_stuff) != 0) {
        perror("atexit()");
        abort();
    }

    mypid = getpid();

    kd = kvm_open(NULL, NULL, NULL, KVM_NO_FILES, "kvm_open");
    if (kd == NULL) {
        perror("kvm_open");
        abort();
    }
}



void *
malloc(size_t size)
{
    void *ptr;

    if (lib == NULL)
        initialize();
    
    ptr = oldmalloc.f(size);

/*
    if (ptr != NULL && maxalloc < size)
        maxalloc = size;
*/
    update_stats();

    return ptr;
}



void *
calloc(size_t num, size_t size)
{
    void *ptr;

    if (lib == NULL)
        initialize();
    
    ptr = oldcalloc.f(num, size);

/*
    if (ptr != NULL && maxalloc < num*size)
        maxalloc = num*size;
*/
    update_stats();

    return ptr;
}



void *
realloc(void *old, size_t size)
{
    void *ptr;

    if (lib == NULL)
        initialize();
    
    ptr = oldrealloc.f(old, size);

/*
    if (ptr != NULL && maxalloc < size)
        maxalloc = size;
*/
    update_stats();

    return ptr;
}
