#!/bin/tcsh
# Datei hashtest.csh: Shell-Skript, um zu bestimmen, wie viel Prozent
# aller Änderungsoperationen auf Hashtafeln dafür gesorgt haben, dass die
# modifizierte Hashtafel eine durchschnittliche Zeilenbelegung von mehr als
# 2, 3, 4, 5 Elementen hat.
#
# Es wird davon ausgegangen, dass dem Skript ein Dateiname als Argument
# übergeben wird, in welche die Parserausgabe umgeleitet wurde.
#
# Der Parser muss mit dem definierten Makro DEBUG_ALL compiliert worden
# sein. Gemessen wird jeweils nach Ausführen einer der vier Operationen
# node_table_add, node_table_remoce, node_table_pop, node_table_double.
#
# Jede zu berücksichtigende Ausgabezeile des Parsers hat die Form:
# stat: length: <l> item_count: <ic> line_count <lc> relativ: <d>
# Zeilen mit d = nan werden nicht berücksichtigt.
# Diese entstehen, wenn ic = lc = 0, weil die Tabelle nach einer
# Löschoperation keine Elemente mehr enthält.

# Teste, ob ein gültiger Dateiname übergeben wurde:
if ( $#argv != 1 ) then
  echo Falsche Anzahl \($#argv\) von Parametern!
  exit -1
endif
if ( ! -f $1.log ) then
  echo Datei $1.log existiert nicht!
  exit -1
endif

# Nimm aus der angegebenen Datei nur solche Zeilen, die nicht zu einer
# leeren Hashtafel gehören ( $9 != "nan" ).
# Gib alle diese Zeilen aus ( print $0; ) und sortiere sich nach dem
# Wert d (sort -nk 9).
# Aus diesem Zwischenergebnis, zähle, wie oft jeder Wert von d vorkommt,
# und mache daraus eine Tabelle mit den zwei Spalten
# "Häufigkeit" und "d-Wert".
# Speichere diese Tabelle in einer Zwischenergebnis-Datei.
awk '{ if ($9 != "nan") print $0; }' < $1.log | sort -nk 9 | awk '{print $9}' | uniq -c > $1.log.tmp

# Berechne nun für jede Zeile in $1.log.tmp, wie oft d-Werte > 2 auftreten
# (Summe gr) und wie oft d-Werte <= 2 auftreten (Summe lt).
# Gib am Ende (gr / (gr + lt)) * 100 aus.
# Dies gibt an, wieviel Prozent der Änderungsoperationen, die eine
# nichtleere Hashtafel zurücklassen, eine durchschnittliche Zeilenbelegung
# mit mehr als 2 Elementen zurücklassen.
echo 2: > $1.hash
awk 'BEGIN{gr=0; lt=0} { if ($2 > 2.0) gr += $1; else lt += $1; } END { print gr " " lt }' < $1.log.tmp | awk '{ print "(" $1 " / " "(" $1 "+" $2 ")) * 100" }' | bc -l >> $1.hash

# Führe eine Analoge Rechnung für alle d-Werte > 3 durch:
echo 3: >> $1.hash
awk 'BEGIN{gr=0; lt=0} { if ($2 > 3.0) gr += $1; else lt += $1; } END { print gr " " lt }' < $1.log.tmp | awk '{ print "(" $1 " / " "(" $1 "+" $2 ")) * 100" }' | bc -l >> $1.hash

# Führe eine Analoge Rechnung für alle d-Werte > 4 durch:
echo 4: >> $1.hash
awk 'BEGIN{gr=0; lt=0} { if ($2 > 4.0) gr += $1; else lt += $1; } END { print gr " " lt }' < $1.log.tmp | awk '{ print "(" $1 " / " "(" $1 "+" $2 ")) * 100" }' | bc -l >> $1.hash

# Führe eine Analoge Rechnung für alle d-Werte > 5 durch:
echo 5: >> $1.hash
awk 'BEGIN{gr=0; lt=0} { if ($2 > 5.0) gr += $1; else lt += $1; } END { print gr " " lt }' < $1.log.tmp | awk '{ print "(" $1 " / " "(" $1 "+" $2 ")) * 100" }' | bc -l >> $1.hash


