#include <stdlib.h>
#include <stdio.h>

int main (int argc, char *argv[])
{
	int num;
	if (argc > 1)
		num = atoi(argv[1]);
	else
		num = 10;

	/* Universelle Zählvariablen */
	int i, j;

	printf("/* Grammar taken from E. Ukkonen: Upper Bounds on the Size of LR(k) Parsers.\n");
	printf(" * Here: Grammar G_%d.\n", num);
	printf(" */\n");

	printf("%%{\n");
	printf("#include <stdio.h>\n");
	printf("\n");
	printf("int yylex (void); // forward\n");
	printf("\n");
	printf("void\n");
	printf("yyerror (char const *s)\n");
	printf("{\n");
	printf("  fprintf(stderr, \"%%s\\n\", s);\n");
	printf("}\n");
	printf("\n");
	printf("%%}\n\n");

	printf("%%token");
	for (i=1; i<=num; i++) {
		printf(" a%d", i);
	}
	printf("\n");

	printf("%%token");
	for (i=1; i<=num; i++) {
		printf(" b%d", i);
	}
	printf("\n");

	printf("%%start S\n");
	printf("\n%%%%\n\n");

	printf("S :\n");
	printf("    A1 {}\n");
	for (i=2; i<=num; i++) {
		printf("  | A%d {}\n", i);
	}
	printf("  ;\n\n");

	for (i=1; i<=num; i++) {
		printf("A%d :\n", i);
		printf("    a%d B%d {}\n", i, i);
		printf("  | b%d {}\n", i);
		for (j=1; j<=num; j++) {
			if (j != i)
				printf("  | a%d A%d {}\n", j, i);
		}
		printf("  ;\n\n");
	}

	for (i=1; i<=num; i++) {
		printf("B%d :\n", i);
		printf("    b%d {}\n", i);
		for(j=1; j<=num; j++) {
			printf("  | a%d B%d {}\n", j, i);
		}
		printf("  ;\n\n");
	}

	printf("%%%%\n\n");
	printf("int\n");
	printf("main (void)\n");
	printf("{\n");
	printf("  return yyparse();\n");
	printf("}\n");

	return(EXIT_SUCCESS);
}
